REBOL [
	Title: "Uniserve: Async HTTP protocol"
	Author: "SOFTINNOV / Nenad Rakocevic"
	Version: 0.9.2
	Date: 27/12/2002
]

install-service [
	name: 'HTTPd
	port-id: 80
	module: 'CGI
	verbose: 0
	
	resp: mime: htfile: file: cgi: bytes: len: http-code: mh: none
	request: reqfile: method: url: ext: post-bytes: none
	lf: #"^/"
	
	header-end: join crlf crlf
	
	url-chars: complement charset [#"#" #"?" #" "]

	do-task: func [data][shared/do-task data self]	; handy short-cut
	
	prefs: context [
		root: %wwwroot/								; path must end with a slash !
		default: "index.html"						; default file for folder request 
		mime-types: load %httpd/mime.types			; mime type list
		dyn-types: [cgi asp]						; extensions that trigger bg processing
	]
	
	err-base: "<html><strong><h1>HTTP Error</h1></strong><h4><p><p>Uniserve/HTTPd service error "
	not-found: join err-base "404<br>File not found</html>"
	bad-perms: join err-base "400<br>You do not have permission to view the file</html>"
	int-error: join err-base "500<br>Internal Server error</html>"
	script-error: join err-base "501<br>Script processing error</html>"

	cgi-data: [prefs/root request client/remote-ip]	; + POSTed data if any
	
	responses: [
		200 "200 OK"
		204 "204 No Content"
		302 "302 Found"
		304 "304 Not Modified"
		400 "400 Forbidden"
		404 "404 Not Found"
		500 "500 Internal Server Error"
	]
			
	make-header: func [code type data /cgi /status line /local][
		out: make string! 1024
		insert out "HTTP/1.0 "
		insert tail out either status [line][select responses code]
		insert tail out "^M^/Server: Uniserve/HTTPd^M^/Connection: close^M^/"
		if code = 302 [insert tail out rejoin ["Location: " data crlf]]
		if code = 200 [
			if not cgi [insert tail out rejoin ["Content-Type: " type crlf]]
			insert tail out rejoin ["Content-Length: " any [all [file? data size? data] length? data] crlf]
		]
		if not cgi [insert tail out crlf]
		if verbose > 1 [log/info ["response headers=>^/" out]]
		out
	]
	
	on-new-client: does [
		stop-at: header-end
	]
	
	on-received: func [data][
		either integer? stop-at [
			; --- POST data received ---
			do-task mold append client/user-data data
			stop-at: header-end		
		][
			; --- new request received ---
			post-bytes: none
			if verbose > 1 [log/info ["new request: =>^/" to-string data]]
			parse/all request: copy data [
				copy method to #" " skip copy url [some url-chars]
				[thru "Content-length:" copy post-bytes to #"^/" | none] to end
			]
			if verbose > 0 [log/info ["method: " method " - resource: " url]]
			if not all [method url][
				log/error "not a valid HTTP request"
				close-client
			]
			reqfile: none
			either all [
				slash <> last url
				ext: find/tail reqfile: last parse url "/" "." 
				not empty? copy ext
			][
				ext: to word! ext					; file extension
			][
				if any [not reqfile empty? reqfile][append url prefs/default]
				ext: 'html
			]
			if slash = first url [remove url]
			if dir? file: join prefs/root url [
				file: rejoin [prefs/root url %/ prefs/default]
				ext: 'html
			]
			either find prefs/dyn-types ext [
				; --- CGI request ---
				either post-bytes [
					client/user-data: reduce cgi-data	; save ctx for remote job
					stop-at: to-integer trim post-bytes	; wait for POSTed data
				][
					do-task remold cgi-data				; execute the cgi in bg
					stop-at: header-end
				]
			][
				; --- Static file request ---
				mime: either mime: find prefs/mime-types ext [first find/reverse mime path!][
					'application/download
				]		
				http-code: 200
				if verbose > 0 [log/info ["local target: " file]]
				file: load mold file	; workaround for an REBOL internal bug
				any [
					all [
						not exists? file
						http-code: 404 mime: "text/html" file: not-found
					]
					all [
						error? try [len: size? file]
						http-code: 400 mime: "text/html" file: bad-perms
					]
					all [len < 10000 file: read/binary file]
				]
				write-client make-header http-code mold :mime file
				write-client file
				close-client
			]
		]
	]
	
	on-task-done: func [data /local valid out value ctype body headers s e][
		either "HTTP" = copy/part data 4 [
			;--- Non Parsed Header output ---
			write-client data
		][
			;--- Parsed Header output ---
			out: ctype: line: none
			valid: parse/all data [
				s: some [
					[
						"Location:" copy value to lf (out: make-header 302 none trim value) |
						"Content-Type:" copy ctype to lf |
						s: "Status:" copy line to lf e: | to lf
					] lf [0 1 cr lf body: to end | none]
				]
			]
			any [
				all [
					not valid
					write-client append make-header 200 "text/html" script-error script-error
				]
				all [out write-client out]
				all [
					ctype
					write-client make-header/cgi 200 trim ctype body
					write-client copy data
				]
				all [
					line
					remove/part data s e
					write-client append make-header/status none none none line data
				]
			]
			
		]
		close-client
	]
	
	on-task-failed: func [reason][
		write-client make-header 500 "text/html" int-error
		write-client int-error
		close-client
	]
	
	stop-at: header-end	; wait for a http request header
]
